"""
dariah.api
~~~~~~~~~~

This module implements the high-level API.
"""

from pathlib import Path

import cophi

from dariah.core import LDA, Vis


def topics(directory, stopwords, num_topics, num_iterations, **kwargs):
    """Train a topic model.

    Parameters:
        directory (str): Path to corpus directory.
        stopwords (int, list): Either a threshold for most frequent words,
            or a list of stopwords.
        num_topics (int): Number of topics.
        num_iterations (int): Number of iterations.
        alpha (float): TODO. Defaults to 0.1.
        eta (float): TODO. Defaults to 0.01.
        random_state (int): TODO. Defaults to None.
        mallet (str): TODO. Defaults to None.

    Returns:
        A topic model and its visualizations.
    """
    # Construct a corpus object:
    corpus = cophi.corpus(directory, metadata=False)
    # Get stopwords and hapax legomena:
    if isinstance(stopwords, int):
        stopwords = corpus.mfw(stopwords)
    stopwords = stopwords + corpus.hapax
    # Remove them from corpus:
    dtm = corpus.drop(corpus.dtm, stopwords)
    # Construct LDA object:
    model = LDA(num_topics=num_topics, num_iterations=num_iterations, **kwargs)
    # Train model:
    model.fit(dtm)
    # Construct visualization object:
    vis = Vis(model)
    return model, vis



"""
dariah.topics.modeling
~~~~~~~~~~~~~~~~~~~~~~

This module implements low-level LDA modeling functions.
"""

from pathlib import Path
import tempfile
import os
import logging
import multiprocessing
import shutil
from typing import Optional, Union

import cophi
import lda
import numpy as np
import pandas as pd

from dariah.mallet import MALLET
from dariah.core import utils


logging.getLogger("lda").setLevel(logging.WARNING)


class LDA:
    """Latent Dirichlet allocation.

    Args:
        num_topics: The number of topics.
        num_iterations: The number of iterations.
        alpha:
        eta:
        random_state:
        mallet:
    """

    def __init__(
        self,
        num_topics: int,
        num_iterations: int = 1000,
        alpha: float = 0.1,
        eta: float = 0.01,
        random_state: int = None,
        mallet: Optional[Union[str, Path]] = None,
    ) -> None:
        self.num_topics = num_topics
        self.num_iterations = num_iterations
        self.alpha = alpha
        self.eta = eta
        self.random_state = random_state
        self.mallet = mallet
        if mallet:
            if not Path(self.mallet).exists():
                # Check if MALLET is in environment variable:
                if not os.environ.get(self.mallet):
                    raise OSError(
                        "MALLET executable was not found. "
                        "'{}' does not exist".format(self.mallet)
                    )
                self.mallet = os.environ.get(self.mallet)
            if not Path(self.mallet).is_file():
                raise OSError(
                    "'{}' is not a file. "
                    "Point to the 'mallet/bin/mallet' file.".format(self.mallet)
                )
        else:
            self._model = lda.LDA(
                n_topics=self.num_topics,
                n_iter=self.num_iterations,
                alpha=self.alpha,
                eta=self.eta,
                random_state=self.random_state,
            )

    def fit(self, dtm: pd.DataFrame) -> None:
        """Fit the model.

        Args:
            dtm: The document-term matrix.
        """
        self._vocabulary = list(dtm.columns)
        self._documents = list(dtm.index)
        dtm = dtm.fillna(0).astype(int)
        if self.mallet:
            self._mallet_lda(dtm)
        else:
            self._riddell_lda(dtm.values)

    @property
    def topics(self):
        """Topics with 200 top words.
        """
        if self.mallet:
            return self._mallet_topics
        else:
            return self._riddell_topics

    @property
    def topic_word(self):
        """Topic-word distributions.
        """
        if self.mallet:
            return self._mallet_topic_word
        else:
            return self._riddell_topic_word

    @property
    def topic_document(self):
        """Topic-document distributions.
        """
        if self.mallet:
            return self._mallet_topic_document
        else:
            return self._riddell_topic_document

    @property
    def topic_similarities(self):
        """Topic similarity matrix.
        """
        data = self.topic_document.T.copy()
        return self._similarities(data)

    @property
    def document_similarities(self):
        """Document similarity matrix.
        """
        data = self.topic_document.copy()
        return self._similarities(data)

    @staticmethod
    def _similarities(data: pd.DataFrame) -> pd.DataFrame:
        """Calculate cosine simliarity matrix.

        Args:
            data: A matrix to calculate similarities for.

        Returns:
            A similarity matrix.
        """
        descriptors = data.columns
        d = data.T @ data
        norm = (data * data).sum(0) ** 0.5
        similarities = d / norm / norm.T
        return pd.DataFrame(similarities, index=descriptors, columns=descriptors)

    def _riddell_lda(self, dtm: pd.DataFrame) -> None:
        """Fit the Riddell LDA model.

        Args:
            dtm: The document-term matrix.
        """
        self._model.fit(dtm)

    @property
    def _riddell_topics(self):
        """Topics of the Riddell LDA model.
        """
        maximum = len(self._vocabulary)
        num_words = 200 if maximum > 200 else maximum
        index = [f"topic{n}" for n in range(self.num_topics)]
        columns = [f"word{n}" for n in range(num_words)]
        topics = [
            np.array(self._vocabulary)[np.argsort(dist)][: -num_words - 1 : -1]
            for dist in self._model.topic_word_
        ]
        return pd.DataFrame(topics, index=index, columns=columns)

    @property
    def _riddell_topic_word(self):
        """Topic-word distributions for Riddell LDA model.
        """
        index = [f"topic{n}" for n in range(self.num_topics)]
        return pd.DataFrame(
            self._model.topic_word_, index=index, columns=self._vocabulary
        )

    @property
    def _riddell_topic_document(self):
        """Topic-document distributions for Riddell LDA model.
        """
        index = [f"topic{n}" for n in range(self.num_topics)]
        return pd.DataFrame(
            self._model.doc_topic_, index=self._documents, columns=index
        ).T

    def _mallet_lda(self, dtm: pd.DataFrame) -> None:
        """Fit the MALLET LDA model.

        Args:
            dtm: The documen-term matrix.
        """
        # Get number of CPUs for threaded processing:
        cpu = multiprocessing.cpu_count() - 1

        # Get temporary directory to dump corpus files:
        self._tempdir = Path(tempfile.gettempdir(), "dariah-topics")
        if self._tempdir.exists():
            shutil.rmtree(str(self._tempdir))
        self._tempdir.mkdir()

        # Export document-term matrix to plaintext files:
        corpus_sequence = Path(self._tempdir, "corpus.sequence")
        cophi.text.utils.export(dtm, corpus_sequence, "plaintext")

        # Construct MALLET object:
        mallet = MALLET(self.mallet)

        # Create a MALLET corpus file:
        corpus_mallet = Path(self._tempdir, "corpus.mallet")
        mallet.import_file(
            input=str(corpus_sequence), output=str(corpus_mallet), keep_sequence=True
        )

        # Construct paths to MALLET output files:
        self._topic_document_file = Path(self._tempdir, "topic-document.txt")
        self._topic_word_file = Path(self._tempdir, "topic-word.txt")
        self._topics_file = Path(self._tempdir, "topics.txt")

        # Train topics:
        mallet.train_topics(
            input=str(corpus_mallet),
            num_topics=self.num_topics,
            num_iterations=self.num_iterations,
            output_doc_topics=self._topic_document_file,
            output_topic_keys=self._topics_file,
            topic_word_weights_file=self._topic_word_file,
            alpha=self.alpha,
            beta=self.eta,
            num_top_words=200,
            num_threads=cpu,
            random_seed=self.random_state,
        )

    @property
    def _mallet_topics(self):
        """Topics of MALLET LDA model.
        """
        maximum = len(self._vocabulary)
        num_words = 200 if maximum > 200 else maximum
        index = [f"topic{n}" for n in range(self.num_topics)]
        columns = [f"word{n}" for n in range(num_words)]
        topics = utils.read_mallet_topics(self._topics_file, num_words)
        return pd.DataFrame(topics, index=index, columns=columns)

    @property
    def _mallet_topic_word(self):
        """Topic-word distributions of MALLET LDA model.
        """
        index = [f"topic{n}" for n in range(self.num_topics)]
        data = pd.read_csv(self._topic_word_file, sep="\t", header=None).dropna()
        data = data.pivot(index=0, columns=1, values=2)
        data.columns.name = None
        data.index.name = None
        data.index = index
        return data

    @property
    def _mallet_topic_document(self):
        """Topic-document distributions of MALLET LDA model.
        """
        data = pd.read_csv(self._topic_document_file, sep="\t", header=None)
        columns = [f"topic{n}" for n in range(self.num_topics)]
        index = data[1]
        data = data.drop([0, 1], axis=1)
        data.columns = list(columns)
        data.index = index
        return data.T

    def __repr__(self):
        return (
            f"<Model: LDA, "
            f"{self.num_topics} topics, "
            f"{self.num_iterations} iterations, "
            f"alpha={self.alpha}, "
            f"eta={self.eta}>"
        )



"""
dariah.topics.visualization
~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import seaborn as sns


class Vis:
    def __init__(self, model):
        self._model = model
        self.topic_document_ = model.topic_document.copy()
        self.topic_word_ = model.topic_word.copy()
        self.document_similarities_ = model.document_similarities.copy()
        self.topic_similarities_ = model.topic_similarities.copy()

    def topic_document(self, cmap="Blues", annot=False, fmt=".2g", cbar=True, **kwargs):
        fig, ax = plt.subplots(**kwargs)
        topic_document_ = self.topic_document_.reindex(
            sorted(self.topic_document_.columns), axis=1
        )
        if topic_document_.shape[0] < topic_document_.shape[1]:
            topic_document_ = topic_document_.T
        sns.heatmap(
            topic_document_,
            linewidth=0.5,
            ax=ax,
            cmap=cmap,
            annot=annot,
            fmt=fmt,
            cbar=cbar,
        )
        return ax

    def topic_word(
        self, words, cmap="Blues", annot=False, fmt=".2g", cbar=True, **kwargs
    ):
        fig, ax = plt.subplots(**kwargs)
        topic_word_ = self.topic_word_.copy().loc[:, words]
        if topic_word_.shape[0] < topic_word_.shape[1]:
            topic_word_ = topic_word_.T
        sns.heatmap(
            topic_word_,
            linewidth=0.5,
            ax=ax,
            cmap=cmap,
            annot=annot,
            fmt=fmt,
            cbar=cbar,
        )

    def topic(self, name, num_words=10, color="grey", **kwargs):
        fig, ax = plt.subplots(**kwargs)
        data = self.topic_word_.loc[name, :].sort_values(ascending=False)
        data = data[:num_words].sort_values()
        data.plot.barh(ax=ax, color=color)
        return ax

    def document(self, name, color="grey", **kwargs):
        fig, ax = plt.subplots(**kwargs)
        data = self.topic_document_.loc[:, name].sort_values()
        data.plot.barh(ax=ax, color=color)
        return ax

    def document_similarities(
        self, cmap="Blues", annot=False, fmt=".2g", cbar=True, **kwargs
    ):
        fig, ax = plt.subplots(**kwargs)
        sns.heatmap(
            self.document_similarities_,
            linewidth=0.5,
            ax=ax,
            cmap=cmap,
            annot=annot,
            fmt=fmt,
            cbar=cbar,
        )
        return ax

    def document_similarities(
        self, cmap="Blues", annot=False, fmt=".2g", cbar=True, **kwargs
    ):
        fig, ax = plt.subplots(**kwargs)
        sns.heatmap(
            self.document_similarities_,
            linewidth=0.5,
            ax=ax,
            cmap=cmap,
            annot=annot,
            fmt=fmt,
            cbar=cbar,
        )
        return ax

    def topic_similarities(
        self, cmap="Blues", annot=False, fmt=".2g", cbar=True, **kwargs
    ):
        fig, ax = plt.subplots(**kwargs)
        sns.heatmap(
            self.topic_similarities_,
            linewidth=0.5,
            ax=ax,
            cmap=cmap,
            annot=annot,
            fmt=fmt,
            cbar=cbar,
        )
        return ax

    def __repr__(self):
        return (
            f"<Visualization: LDA, "
            f"{self._model.num_topics} topics, "
            f"{self._model.num_iterations} iterations, "
            f"alpha={self._model.alpha}, "
            f"eta={self._model.eta}>"
        )



"""
dariah.utils
~~~~~~~~~~~~

This module implements general helper functions.
"""

import logging
import subprocess


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


def call(args: list) -> bool:
    """Call a subprocess.

    Parameter:
        args (list): The subprocess’ arguments.

    Returns:
        True, if call was successful.
    """
    for message in _process(args):
        if len(message) < 100 and len(message) > 5:
            logger.info(message)
    return True


def _process(args: list):
    """Construct a process object.
    """
    popen = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )
    # Yield every line of stdout:
    for line in iter(popen.stderr.readline, ""):
        yield line.strip()
    popen.stdout.close()
    code = popen.wait()
    if code:
        raise subprocess.CalledProcessError(code, args)



"""
dariah.mallet.core
~~~~~~~~~~~~~~~~~~

This module implements the core functions of the MALLET sub-package.
"""

from dariah.mallet import utils


def call(command, executable, **parameters):
    """Call MALLET.

    Parameter:
        command (str): Command for MALLET.
        executable (str): Path to MALLET executable.
        **parameter: Additional parameters for MALLET.

    Returns:
        True, if call was successful.
    """
    # Basic subprocess command:
    args = [executable, command]

    # Append additional parameters:
    for parameter, value in parameters.items():
        # Support synonyms for `-input` parameter:
        if parameter in {"filepath", "directory", "path", "corpus"}:
            args.append("--input")
        elif parameter in {"random_seed", "random_state"}:
            if value:
                args.append("--random-seed")
        else:
            args.append("--{}".format(parameter.replace("_", "-")))
        if value and value != True:
            args.append(str(value))
    return utils.call(args)



"""
dariah.mallet.api
~~~~~~~~~~~~~~~~

This module implements the high-level API to communicate with
the CLI interface of MALLET.
"""

from dariah.mallet import core


class MALLET:
    """Machine Learning for Language Toolkit (MALLET).
    """

    def __init__(self, executable):
        self.executable = executable

    def import_dir(self, **parameters):
        """Load contents of a directory into MALLET instances.
        """
        return core.call("import-dir", self.executable, **parameters)

    def import_file(self, **parameters):
        """Load a file into MALLET instances.
        """
        return core.call("import-file", self.executable, **parameters)

    def import_svmlight(self, **parameters):
        """Load SVMLight data files into MALLET instances.
        """
        return core.call("import-svmlight", self.executable, **parameters)

    def info(self, **parameters):
        """Get information about MALLET instances.
        """
        return core.call("info", self.executable, **parameters)

    def train_classifier(self, **parameters):
        """Train a classifier from MALLET data files.
        """
        return core.call("train-classifier", self.executable, **parameters)

    def classify_dir(self, **parameters):
        """Classify the contents of a directory with a saved classifier.
        """
        return core.call("classify-dir", self.executable, **parameters)

    def classify_file(self, **parameters):
        """Classify data from a single file with a saved classifier.
        """
        return core.call("classify-file", self.executable, **parameters)

    def classify_svmlight(self, **parameters):
        """Classify data from a single file in SVMLight format.
        """
        return core.call("classify-svmlight", self.executable, **parameters)

    def train_topics(self, **parameters):
        """Train a topic model from MALLET data files.
        """
        return core.call("train-topics", self.executable, **parameters)

    def infer_topics(self, **parameters):
        """Use a trained topic model to infer topics for new documents.
        """
        return core.call("infer-topics", self.executable, **parameters)

    def evaluate_topics(self, **parameters):
        """Estimate the probability of new documents under a trained model.
        """
        return core.call("evaluate-topics", self.executable, **parameters)

    def prune(self, **parameters):
        """Remove features based on frequency or information gain.
        """
        return core.call("prune", self.executable, **parameters)

    def split(self, **parameters):
        """Divide data into testing, training, and validation portions.
        """
        return core.call("split", self.executable, **parameters)

    def bulk_load(self, **parameters):
        """For big input files, efficiently prune vocabulary and import docs.
        """
        return core.call("bulk-load", self.executable, **parameters)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dariah', 'dariah.core', 'dariah.mallet']

package_data = \
{'': ['*']}

install_requires = \
['cophi>=1.3.2,<2.0.0',
 'lda>=1.1.0,<2.0.0',
 'matplotlib>=3.0.3,<4.0.0',
 'numpy>=1.16.2,<2.0.0',
 'pandas>=0.24.2,<0.25.0',
 'regex>=2019.4.12,<2020.0.0',
 'seaborn>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'dariah',
    'version': '2.0.1',
    'description': 'A library for topic modeling and visualization.',
    'long_description': 'A library for topic modeling and visualization\n==============================================\n\nDARIAH Topics is an easy-to-use Python library for topic modeling and visualization. Getting started is `really easy`. All you have to do is import the library – you can train a model straightaway from raw textfiles.\n\nIt supports two implementations of `latent Dirichlet allocation <http://www.jmlr.org/papers/volume3/blei03a/blei03a.pdf>`_:\n\n- The lightweight, Cython-based package `lda <https://pypi.org/project/lda/>`_\n- The more robust, Java-based package `MALLET <http://mallet.cs.umass.edu/topics.php>`_\n\n\nInstallation\n------------\n\n::\n\n    $ pip install dariah\n\n\nExample\n-------\n\n>>> import dariah\n>>> dariah.topics(directory="british-fiction-corpus",\n...               stopwords=100,\n...               num_topics=10,\n...               num_iterations=1000)\n\n\nDeveloping\n----------\n\n`Poetry <https://poetry.eustace.io/>`_ automatically creates a virtual environment, builds and publishes the project to `PyPI <https://pypi.org/>`_. Install dependencies with:\n\n::\n\n    $ poetry install\n\nrun tests:\n\n::\n\n    $ poetry run pytest\n\n\nformat code:\n\n::\n\n    $ poetry run black dariah\n\n\nbuild the project:\n\n::\n\n    $ poetry build\n\n\nand publish it on `PyPI <https://pypi.org/>`_:\n\n::\n\n    $ poetry publish\n\n\nAbout DARIAH-DE\n---------------\n\n`DARIAH-DE <https://de.dariah.eu>`_ supports research in the humanities and cultural sciences with digital methods and procedures. The research infrastructure of DARIAH-DE consists of four pillars: teaching, research, research data and technical components. As a partner in `DARIAH-EU <http://dariah.eu/>`_, DARIAH-DE helps to bundle and network state-of-the-art activities of the digital humanities. Scientists use DARIAH, for example, to make research data available across Europe. The exchange of knowledge and expertise is thus promoted across disciplines and the possibility of discovering new scientific discourses is encouraged.\n\nThis software library has been developed with support from the DARIAH-DE initiative, the German branch of DARIAH-EU, the European Digital Research Infrastructure for the Arts and Humanities consortium. Funding has been provided by the German Federal Ministry for Research and Education (BMBF) under the identifier 01UG1610J.\n\n.. image:: https://raw.githubusercontent.com/DARIAH-DE/Topics/master/docs/images/dariah-de_logo.png\n.. image:: https://raw.githubusercontent.com/DARIAH-DE/Topics/master/docs/images/bmbf_logo.png\n',
    'author': 'DARIAH-DE',
    'author_email': None,
    'url': 'https://dariah-de.github.io/Topics',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
This is a library for topic modeling and visualization.

Using the high-level API is easy:

```
>>> model, vis = dariah.topics(directory="corpus",
                               stopwords=100,
                               num_topics=10,
                               num_iterations=1000)
```

"""

from dariah.api import topics



"""
dariah.topics.utils
~~~~~~~~~~~~~~~~~~~

This module implements helper functions for topic modeling.
"""

from typing import Generator, List
from pathlib import Path

import cophi


def read_mallet_topics(path: Path, num_words: int) -> Generator[List[str], None, None]:
    """Read a MALLET topics file.

    Args:
        path: Filepath to the topics file.
        num_words: Number of words for a topic.

    Yields:
        A list of tokens, i.e. a topic.
    """
    with path.open("r", encoding="utf-8") as file:
        for row in file:
            sequence = row.split("\t")[2]
            yield list(cophi.text.utils.find_tokens(sequence))[:200]
